import unittest

class ZODBEvolutionManagerTests(unittest.TestCase):
    def _evolve(self, *arg, **kw):
        from repoze.evolution import evolve_to_latest
        return evolve_to_latest(*arg, **kw)

    def _getTargetClass(self):
        from repoze.evolution import ZODBEvolutionManager
        return ZODBEvolutionManager

    def _makeOne(self, root, sw_version):
        klass = self._getTargetClass()
        context = DummyPersistent(root)
        manager = klass(context, 'repoze.evolution.tests.fixtureapp.evolve',
                        sw_version)
        manager.transaction = DummyTransaction()
        return manager

    def test_verify(self):
        from repoze.evolution import IEvolutionManager
        from zope.interface.verify import verifyClass
        from zope.interface.verify import verifyObject
        klass = self._getTargetClass()
        verifyClass(IEvolutionManager, klass)
        inst = klass(None, None, None)
        verifyObject(IEvolutionManager, inst)
        
    def test_success_no_db_version(self):
        root = {}
        manager = self._makeOne(root, 1)
        version = self._evolve(manager)
        self.assertEqual(version, 1)
        self.assertEqual(manager.context.evolved, None)
        self.assertEqual(manager.transaction.committed, 1)
        reg = root['repoze.evolution']
        self.assertEqual(reg['repoze.evolution.tests.fixtureapp.evolve'], 1)

    def test_success_with_db_version(self):
        root = {'repoze.evolution':
                {'repoze.evolution.tests.fixtureapp.evolve':1}}
        manager = self._makeOne(root, 2)
        version = self._evolve(manager)
        self.assertEqual(version, 2)
        self.assertEqual(manager.context.evolved, 2)
        self.assertEqual(manager.transaction.committed, 1)
        reg = root['repoze.evolution']
        self.assertEqual(reg['repoze.evolution.tests.fixtureapp.evolve'], 2)

    def test_evolve_error(self):
        root = {'repoze.evolution':
                {'repoze.evolution.tests.fixtureapp.evolve':0}}
        manager = self._makeOne(root, 3)
        self.assertRaises(ValueError, self._evolve, manager)
        self.assertEqual(manager.context.evolved, 2)
        self.assertEqual(manager.transaction.committed, 2)
        reg = root['repoze.evolution']
        self.assertEqual(reg['repoze.evolution.tests.fixtureapp.evolve'], 2)

    def test_noevolve(self):
        root = {'repoze.evolution':
                {'repoze.evolution.tests.fixtureapp.evolve':1}}
        manager = self._makeOne(root, 1)
        version = self._evolve(manager)
        self.assertEqual(version, 1)
        self.assertEqual(manager.context.evolved, None)
        self.assertEqual(manager.transaction.committed, 0)
        reg = root['repoze.evolution']
        self.assertEqual(reg['repoze.evolution.tests.fixtureapp.evolve'], 1)

    def test_evolve_swver_not_integer(self):
        root = {'repoze.evolution':
                {'repoze.evolution.tests.fixtureapp.evolve':1}}
        manager = self._makeOne(root, '1')
        self.assertRaises(ValueError, self._evolve, manager)

    def test_evolve_dbver_not_integer(self):
        root = {'repoze.evolution':
                {'repoze.evolution.tests.fixtureapp.evolve':'1'}}
        manager = self._makeOne(root, 1)
        self.assertRaises(ValueError, self._evolve, manager)

class Dummy(object):
    pass

class DummyPersistent(object):
    evolved = None
    def __init__(self, root):
        self._p_jar = Dummy()
        self._p_jar.root = lambda *arg: root
        self.error = False
        
class DummyTransaction(object):
    committed = 0
    begun = 0
    def commit(self):
        self.committed = self.committed + 1

    def begin(self):
        self.begun = self.begun + 1
        
        
        



from pkg_resources import EntryPoint
from zope.interface import implements
from zope.interface import Interface

class IEvolutionManager(Interface):
    def get_sw_version():
        """ Return the software version of the managed package """

    def get_db_version():
        """ Return the database version of the managed package """

    def evolve_to(version):
        """ Perform work to evolve to the integer ``version``.  This
        method is also responsible for setting the db version after a
        success."""

class ZODBEvolutionManager:
    key = 'repoze.evolution'
    implements(IEvolutionManager)
    def __init__(self, context, evolve_packagename, sw_version):
        """ Initialize a ZODB evolution manager.  ``context`` is an
        object which must inherit from ``persistent.Persistent`` that
        will be passed in to each evolution step.  evolve_packagename
        is the Python dotted package name of a package which contains
        evolution scripts.  ``sw_version`` is the current software
        version of the software represented by this manager."""
        import transaction
        self.transaction = transaction
        self.context = context
        self.package_name = evolve_packagename
        self.sw_version = sw_version

    @property
    def root(self):
        return self.context._p_jar.root()

    def get_sw_version(self):
        return self.sw_version

    def get_db_version(self):
        registry = self.root.setdefault(self.key, {})
        db_version = registry.get(self.package_name)
        if db_version is None:
            self.transaction.begin()
            self._set_db_version(self.sw_version)
            self.transaction.commit()
            db_version = self.sw_version
        return db_version

    def evolve_to(self, version):
        scriptname = '%s.evolve%s' % (self.package_name, version)
        evmodule = EntryPoint.parse('x=%s' % scriptname).load(False)
        self.transaction.begin()
        evmodule.evolve(self.context)
        self._set_db_version(version)
        self.transaction.commit()

    def _set_db_version(self, version):
        registry = self.root.setdefault(self.key, {})
        db_version = registry[self.package_name] = version
        self.root[self.key] = registry

def evolve_to_latest(manager):
    """ Evolve the database to the latest software version using the
    ``manager`` object.  """
    db_version = manager.get_db_version()
    sw_version = manager.get_sw_version()
    if not isinstance(sw_version, int):
        raise ValueError('software version %s is not an integer' %
                         sw_version)
    if not isinstance(db_version, int):
        raise ValueError('database version %s is not an integer' %
                         db_version)
    if db_version < sw_version:
        for version in range(db_version+1, sw_version+1):
            manager.evolve_to(version)
        return version
    return db_version

